#############################################################################
##
##  PackageInfo.g for the GAP 4 package AtlasRep                Thomas Breuer
##
SetPackageInfo( rec(
PackageName :=
  "AtlasRep",
MyVersion :=
  "1r2p1",
MyWWWHome :=
  "http://www.math.rwth-aachen.de/~Thomas.Breuer",
Subtitle :=
  "A GAP Interface to the Atlas of Group Representations",
Version :=
  JoinStringsWithSeparator( SplitString( ~.MyVersion, "rp" ), "." ),
Autoload :=
  false,
Date :=
  # "03/04/2001" -- Version 1.0
  # "23/10/2002" -- Version 1.1
  # "06/11/2003" -- Version 1.2
  "05/04/2004",
PackageWWWHome :=
  Concatenation( ~.MyWWWHome, "/", LowercaseString( ~.PackageName ) ),
ArchiveURL :=
  Concatenation( ~.PackageWWWHome, "/", LowercaseString( ~.PackageName ),
                 ~.MyVersion ),
ArchiveFormats :=
  ".tar.gz,.zoo",
Persons := [
  rec(
    LastName := "Wilson",
    FirstNames := "Robert A.",
    IsAuthor := true,
    IsMaintainer := true,
    Email := "R.A.Wilson@bham.ac.uk",
    WWWHome := "http://for.mat.bham.ac.uk/R.A.Wilson",
    Place := "Birmingham",
    Institution := Concatenation( [
      "School of Mathematics and Statistics, ",
      "The University of Birmingham",
      ] )
    ),
  rec(
    LastName := "Parker",
    FirstNames := "Richard A.",
    IsAuthor := true,
    IsMaintainer := true,
    Email := "richard@ukonline.co.uk",
    WWWHome := "http://www.ukonline.co.uk/richard"
  ),
  rec(
    LastName := "Bray",
    FirstNames := "John",
    IsAuthor := true,
    IsMaintainer := true,
    Email := "jnb@for.mat.bham.ac.uk",
    WWWHome := "http://web.mat.bham.ac.uk/J.N.Bray"
  ),
  rec(
    LastName := "Breuer",
    FirstNames := "Thomas",
    IsAuthor := true,
    IsMaintainer := true,
    Email := "sam@math.rwth-aachen.de",
    WWWHome := ~.MyWWWHome,
    Place := "Aachen",
    Institution := "Lehrstuhl D f&uuml;r Mathematik, RWTH Aachen",
    PostalAddress := Concatenation( [
      "Thomas Breuer\n",
      "Lehrstuhl D f&uuml;r Mathematik\n",
      "Templergraben 64\n",
      "52062 Aachen\n",
      "Germany"
      ] )
  ),
  ],
Status :=
  "accepted",
CommunicatedBy :=
  "Herbert Pahlings (Aachen)",
AcceptDate :=
  "04/2001",
README_URL :=
  Concatenation( ~.PackageWWWHome, "/README" ),
PackageInfoURL :=
  Concatenation( ~.PackageWWWHome, "/PackageInfo.g" ),
AbstractHTML := Concatenation( [
  "The package provides a <span class=\"pkgname\">GAP</span> interface ",
  "to the <a href=\"http://web.mat.bham.ac.uk/atlas/v2.0\">",
  "Atlas of Group Representations</a>"
  ] ),
PackageDoc := rec(
  BookName :=
    "AtlasRep",
  ArchiveURLSubset :=
    [ "doc", "htm" ],
  HTMLStart :=
    "htm/chapters.htm",
  PDFFile :=
    "doc/manual.pdf",
  SixFile :=
    "doc/manual.six",
  LongTitle :=
    "An Atlas of Group Representations",
  Autoload :=
    true
  ),
Dependencies := rec(
  GAP :=
    ">= 4.4",
  NeededOtherPackages :=
    [], # [["gpisotyp", ">= 1.0"]],
  SuggestedOtherPackages :=
    [ [ "ctbllib", ">= 1.0" ], [ "tomlib", ">= 1.0" ] ],
  # needed external conditions (programs, operating system, ...)  provide
  # just strings as text or
  # pairs [text, URL] where URL  provides further information
  # about that point.
  # (no automatic test will be done for this, do this in your
  # 'AvailabilityTest' function below)
  ExternalConditions :=
    []
  ),
AvailabilityTest :=
  function()
  local sysproglist, name, sysprog;

  # Test for the existence of the following programs.
  # (The package works without these programs if all data files
  # are contained in the local installation, so we always return `true',
  # just some warnings are issued.)
  sysproglist:= [ "date", "gzip", "ls", "perl", "wget" ];
  name:= Filename( DirectoriesPackageLibrary( "atlasrep", "etc" ),
                   "perlftp.pl" );
  if name = fail then
    Info( InfoWarning, 1,
          "Package `atlasrep':  script `etc/perlftp.pl' not available");
  elif not IsExecutableFile( name ) then
    Info( InfoWarning, 1,
          "Package `atlasrep':  script `etc/perlftp.pl' not executable");
  fi;
  for sysprog in sysproglist do
    name:= Filename( DirectoriesSystemPrograms(), sysprog );
    if name = fail then
      Info( InfoWarning, 1,
            "Package `atlasrep':  The system program `", sysprog,
            "' is not available." );
    elif not IsExecutableFile( name ) then
      Info( InfoWarning, 1,
            "Package `atlasrep':  The system program `", sysprog,
            "' is not executable." );
    fi;
  od;

  return true;
  end,
TestFile :=
  "tst/testall.g",
Keywords :=
  ["group representations", "finite simple groups"]
) );

#############################################################################
##
#E

